#include "exttext.h"

ExtEdit::ExtEdit(QWidget *parent) : QTextEdit(parent)
{
    innerEdit = new QTextEdit(parent);
    innerEdit->setVisible(false);
    m_clipboard = QApplication::clipboard();
    prevInsertedSize = 0;
    m_histSize = 0;
    m_backscan = 0;

    QFont f("unexistent");
    f.setStyleHint(QFont::Monospace);
    f.setPixelSize(11);
    setFont(f);

    setFixedSize(681, 511);

    QFontMetrics fm(f);
    int maxchars = width() / fm.maxWidth();

    QString greeting;
    greeting += QString("Attention! This terminal will show your passwords!\n\n");
    greeting += QString("Also, scrolling back through the history is buggy.\n");

    greeting += QString("Doing so to an invisible 'exit' command, then typing Enter,\n");
    greeting += QString("will equally kill the session... You've been warned.\n");
    greeting += QString("To reduce any relevant side-effects, back-scrolled commands\n");
    greeting += QString("will always be entered Un-edited.\n\n");

    greeting += QString("The maximum terminal width in characters is: ") + QString::number(maxchars - 2);
    greeting += QString("\nThe appearance of the scroll-bar will reduce that by 3.\n\n");

    insertPlainText(greeting);
    m_CursorPos = textCursor().position();
}

ExtEdit::~ExtEdit()
{
    delete innerEdit;
}

void ExtEdit::keyPressEvent(QKeyEvent *pressed)
{
    QString temp;
    QTextCursor cursor = textCursor();
    int old_size = innerEdit->toPlainText().length();

    const QChar newline(QChar('\n'));
    const QChar cr('\r');
#ifdef __linux__
    const QChar eol = newline;
#elif _WIN32
    const QString eol = QString(cr) + QString(newline);
#else
    const QChar eol = cr;
#endif

    m_eol = QString(eol);

    if (pressed->key() == Qt::Key_Left || pressed->key() == Qt::Key_Right)
        return;

    if ((pressed->key() == Qt::Key_C) &&
            (QApplication::keyboardModifiers() & Qt::ControlModifier)) {
        innerEdit->clear();
        emit askSIGINT();
        return;
    }

    if ((pressed->key() == Qt::Key_Z) &&
            (QApplication::keyboardModifiers() & Qt::ControlModifier)) {
        innerEdit->clear();
        emit askSIGSTP();
        return;
    }


    if (pressed->key() == Qt::Key_V &&
            (QApplication::keyboardModifiers() & Qt::ControlModifier) &&
            (QApplication::keyboardModifiers() & Qt::ShiftModifier) &&
            m_histSize == 0) {

        cursor.clearSelection();
        setTextCursor(cursor);
        temp = m_clipboard->text();
        innerEdit->setText(innerEdit->toPlainText() + temp);
        cursor.movePosition(QTextCursor::Right, QTextCursor::MoveAnchor, m_CursorPos);
        cursor.movePosition(QTextCursor::Left, QTextCursor::KeepAnchor, old_size);
        setTextCursor(cursor);
        temp = innerEdit->toPlainText();
        insertPlainText(temp);
        innerEdit->moveCursor(QTextCursor::End);
        return;

    } else if (pressed->key() == Qt::Key_Up && prevInsertedSize == 0) {
        prevInsertedSize = old_size + 1;
        emit forwardText(QString(QChar(16)).remove(eol));
        return;
    } else if (pressed->key() == Qt::Key_Down && prevInsertedSize == 0) {
        prevInsertedSize = old_size + 1;
        emit forwardText(QString(QChar(14)).remove(eol));
        return;

    } else if (m_histSize > 0 && pressed->key() != Qt::Key_Enter && pressed->key() != Qt::Key_Return) {
        return;

    } else {
        ((QObject *)innerEdit)->event(pressed);
        temp = innerEdit->toPlainText();
    }

    if (pressed->key() == Qt::Key_Up || pressed->key() == Qt::Key_Down)
        return;

    cursor.clearSelection();
    setTextCursor(cursor);
    cursor.movePosition(QTextCursor::Right, QTextCursor::MoveAnchor, m_CursorPos);
    cursor.movePosition(QTextCursor::Left, QTextCursor::KeepAnchor, old_size);
    setTextCursor(cursor);

    if (! temp.contains(eol)) {
        insertPlainText(temp);   //  Will also print passwords in plain sight !
    } else {

        innerEdit->clear();
        if (m_histSize == 0) {
            emit forwardText(temp);
        } else {
            m_histSize = 0;
            emit forwardText(QString(eol));
        }
    }
}

void ExtEdit::keyReleaseEvent(QKeyEvent *released)
{
    ((QObject *)innerEdit)->event(released);
}

void ExtEdit::receiveText(QString stream)
{
    QTextCursor cursor = textCursor();

    if (prevInsertedSize > 0) {
        QString stream2 = ExtEdit::sanitize(stream);
        int bs = ExtEdit::backspaces(stream);   //  Computed for the second time.
        int ret = ExtEdit::return_pos(stream);

        if (ret > -1) {

//            moveCursor(QTextCursor::End);
//            moveCursor(QTextCursor::StartOfLine, QTextCursor::KeepAnchor);

            innerEdit->clear();
            innerEdit->setPlainText(stream2.mid(textCursor().positionInBlock() + ret));
            m_backscan = 0;

        } else if (bs > 0) {
            int bsc = bs - innerEdit->toPlainText().length();

            if (bsc < 0)
                bsc = 0;

            bs -= bsc;

            innerEdit->moveCursor(QTextCursor::End);
            cursor = innerEdit->textCursor();
            cursor.movePosition(QTextCursor::Left, QTextCursor::KeepAnchor, bs);
            innerEdit->setTextCursor(cursor);
            innerEdit->insertPlainText(stream2.mid(bsc));

            m_backscan = bsc;

        } else {

            innerEdit->moveCursor(QTextCursor::End);
            innerEdit->moveCursor(QTextCursor::Start, QTextCursor::KeepAnchor);

            innerEdit->insertPlainText(innerEdit->toPlainText() + stream2);
            m_backscan = 0;

        }


        insertPlainText(innerEdit->toPlainText());
        moveCursor(QTextCursor::End);
        cursor = textCursor();
        m_histSize = innerEdit->toPlainText().length();
        cursor.movePosition(QTextCursor::Left, QTextCursor::KeepAnchor, m_histSize);
        setTextCursor(cursor);
        m_CursorPos = cursor.position();

        prevInsertedSize = 0;
        return;
    }

    shortenLog();

    insertPlainText(stream);

    moveCursor(QTextCursor::End);
    m_CursorPos = textCursor().position();
}

QString ExtEdit::sanitize(QString &dirty)
{
    QByteArray accum;
    int i = 0;
    int s = 0;
    int bs = backspaces(dirty);

    for ( i = bs; i < dirty.length(); ++i ) {
        s = dirty[i].unicode();
        if ((s > 31) && (s < 127)) {
            accum.append(s);
        }
    }

    i = accum.length();
    while (true) {
        --i;
        if (i < 0 || accum.at(i) != 32)
            break;
    }

    accum = accum.mid(0, i + 1);

    return QString::fromLocal8Bit(accum);
}

void ExtEdit::shortenLog()
{
    QString *spool1 = new QString;
    QString *spool2 = new QString;
    QStringList fragments;
    *spool1 = toPlainText();
    fragments = spool1->split(m_eol);
    int linecount = fragments.length();

    if (linecount > 500) {
        int i = linecount - 500;
        int j = linecount;

        while (i < j) {
            spool2->append(fragments.at(i++));
            spool2->append(m_eol);
        }

        clear();
        insertPlainText(*spool2);
    }
    delete spool2;
    delete spool1;
}

int ExtEdit::backspaces(QString &stream)
{
    int i = 0;
    int j = stream.length();

    int bs = 8;

    int s = 0;

    while (i < j) {
        s = stream.at(i).unicode();
        if (s != bs) {
            break;
        }
        ++i;
    }

    return i;
}

int ExtEdit::return_pos(QString &stream)
{
    int i = 0;
    int j = stream.length();
    int s = 0;
    int csbef = 0;
    int out = -1;

    for ( ; i < j; ++i) {
        s = stream.at(i).unicode();
        if (s > 31 && s < 127)
            csbef++;
        if (s == 13)
            out = csbef;
    }

    return out;
}

void ExtEdit::mousePressEvent(QMouseEvent *event)
{
    if (m_histSize == 0)
        QTextEdit::mousePressEvent(event);
}

void ExtEdit::mouseDoubleClickEvent(QMouseEvent *event)
{
    if (m_histSize == 0)
        QTextEdit::mouseDoubleClickEvent(event);
}

